from labml import lab
from labml_helpers.datasets.remote import DatasetServer
from torchvision import datasets, transforms


def main():
    transform = transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize((0.1307,), (0.3081,))
    ])

    train_dataset = datasets.MNIST(str(lab.get_data_path()),
                                   train=True,
                                   download=True,
                                   transform=transform)

    valid_dataset = datasets.MNIST(str(lab.get_data_path()),
                                   train=False,
                                   download=True,
                                   transform=transform)

    ds = DatasetServer()
    ds.add_dataset('mnist_train', train_dataset)
    ds.add_dataset('mnist_valid', valid_dataset)

    ds.start()


if __name__ == '__main__':
    main()



import pickle

import uvicorn
from fastapi import FastAPI, Request, Response
from torch.utils.data import Dataset


class _ServerDataset:
    def __init__(self, name: str, dataset: Dataset):
        self.dataset = dataset
        self.name = name

    def len_handler(self, request: Request):
        sample = pickle.dumps(len(self.dataset))
        return Response(sample, media_type='binary/pickle')

    def item_handler(self, request: Request, idx: str):
        sample = self.dataset[int(idx)]

        sample = pickle.dumps(sample)
        return Response(sample, media_type='binary/pickle')


class DatasetServer:
    r"""
    Remote dataset server

    `Here's a sample usage of the server <https://github.com/labmlai/labml/blob/master/helpers/labml_helpers/datasets/remote/test/mnist_server.py>`_
    """

    def __init__(self):
        self.app = FastAPI()
        self.datasets = {}

    def add_dataset(self, name: str, dataset: Dataset):
        """
        Add a dataset


        Arguments:
            name (str): name of the data set
            dataset (Dataset): dataset to be served
        """
        assert name not in self.datasets
        sd = _ServerDataset(name, dataset)
        self.datasets[name] = sd
        self.app.add_api_route("/" + name + "/len", sd.len_handler, methods=["GET"])
        self.app.add_api_route("/" + name + "/item/{idx}", sd.item_handler, methods=["GET"])

    def start(self, host: str = "0.0.0.0", port: int = 8000):
        """
        Start the server

        Arguments:
            host (str): hostname of the server
            port (int): server port
        """
        uvicorn.run(self.app, host=host, port=port)


def _test():
    from labml import lab
    from torchvision import datasets, transforms

    transform = transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize((0.1307,), (0.3081,))
    ])

    dataset = datasets.MNIST(str(lab.get_data_path()),
                             train=True,
                             download=True,
                             transform=transform)
    s = DatasetServer()

    s.add_dataset('mnist_train', dataset)

    s.start()


if __name__ == '__main__':
    _test()



import dataclasses

import torch

from labml import tracker
from . import Metric


@dataclasses.dataclass
class AccuracyState:
    samples: int = 0
    correct: int = 0

    def reset(self):
        self.samples = 0
        self.correct = 0


class Accuracy(Metric):
    data: AccuracyState

    def __init__(self, ignore_index: int = -1):
        super().__init__()
        self.ignore_index = ignore_index

    def __call__(self, output: torch.Tensor, target: torch.Tensor):
        output = output.view(-1, output.shape[-1])
        target = target.view(-1)
        pred = output.argmax(dim=-1)
        mask = target == self.ignore_index
        pred.masked_fill_(mask, self.ignore_index)
        n_masked = mask.sum().item()
        self.data.correct += pred.eq(target).sum().item() - n_masked
        self.data.samples += len(target) - n_masked

    def create_state(self):
        return AccuracyState()

    def set_state(self, data: any):
        self.data = data

    def on_epoch_start(self):
        self.data.reset()

    def on_epoch_end(self):
        self.track()

    def track(self):
        if self.data.samples == 0:
            return
        tracker.add("accuracy.", self.data.correct / self.data.samples)


class AccuracyMovingAvg(Metric):
    def __init__(self, ignore_index: int = -1, queue_size: int = 5):
        super().__init__()
        self.ignore_index = ignore_index
        tracker.set_queue('accuracy.*', queue_size, is_print=True)

    def __call__(self, output: torch.Tensor, target: torch.Tensor):
        output = output.view(-1, output.shape[-1])
        target = target.view(-1)
        pred = output.argmax(dim=-1)
        mask = target == self.ignore_index
        pred.masked_fill_(mask, self.ignore_index)
        n_masked = mask.sum().item()
        if len(target) - n_masked > 0:
            tracker.add('accuracy.', (pred.eq(target).sum().item() - n_masked) / len(target) - n_masked)

    def create_state(self):
        return None

    def set_state(self, data: any):
        pass

    def on_epoch_start(self):
        pass

    def on_epoch_end(self):
        pass


class BinaryAccuracy(Accuracy):
    def __call__(self, output: torch.Tensor, target: torch.Tensor):
        pred = output.view(-1) > 0
        target = target.view(-1)
        self.data.correct += pred.eq(target).sum().item()
        self.data.samples += len(target)


class AccuracyDirect(Accuracy):
    data: AccuracyState

    def __call__(self, output: torch.Tensor, target: torch.Tensor):
        output = output.view(-1)
        target = target.view(-1)
        self.data.correct += output.eq(target).sum().item()
        self.data.samples += len(target)



import torch

from labml.configs import BaseConfigs, hyperparams, option


class DeviceInfo:
    def __init__(self, *,
                 use_cuda: bool,
                 cuda_device: int):
        self.use_cuda = use_cuda
        self.cuda_device = cuda_device
        self.cuda_count = torch.cuda.device_count()

        self.is_cuda = self.use_cuda and torch.cuda.is_available()
        if not self.is_cuda:
            self.device = torch.device('cpu')
        else:
            if self.cuda_device < self.cuda_count:
                self.device = torch.device('cuda', self.cuda_device)
            else:
                self.device = torch.device('cuda', self.cuda_count - 1)

    def __str__(self):
        if not self.is_cuda:
            return "CPU"

        if self.cuda_device < self.cuda_count:
            return f"GPU:{self.cuda_device} - {torch.cuda.get_device_name(self.cuda_device)}"
        else:
            return (f"GPU:{self.cuda_count - 1}({self.cuda_device}) "
                    f"- {torch.cuda.get_device_name(self.cuda_count - 1)}")


class DeviceConfigs(BaseConfigs):
    r"""
    This is a configurable module to get a single device to train model on.
    It can pick up CUDA devices and it will fall back to CPU if they are not available.

    It has other small advantages such as being able to view the
    actual device name on configurations view of
    `labml app <https://github.com/labmlai/labml/tree/master/app>`_

    Arguments:
        cuda_device (int): The CUDA device number. Defaults to ``0``.
        use_cuda (bool): Whether to use CUDA devices. Defaults to ``True``.

    `Here's an example usage <https://github.com/labmlai/labml/blob/master/samples/pytorch/mnist/e_labml_helpers.py>`_.
    """
    cuda_device: int = 0
    use_cuda: bool = True

    device_info: DeviceInfo

    device: torch.device

    def __init__(self):
        super().__init__(_primary='device')


@option(DeviceConfigs.device)
def _device(c: DeviceConfigs):
    return c.device_info.device


hyperparams(DeviceConfigs.cuda_device, DeviceConfigs.use_cuda,
            is_hyperparam=False)


@option(DeviceConfigs.device_info)
def _device_info(c: DeviceConfigs):
    return DeviceInfo(use_cuda=c.use_cuda,
                      cuda_device=c.cuda_device)



import pickle

import matplotlib.pyplot as plt
import urllib3
from torch.utils.data import Dataset


class RemoteDataset(Dataset):
    """
    Remote dataset

    Arguments:
        name (str): name of the data set, as specified in
         :class:`labml_helpers.datasets.remote.DatasetServer`
        host (str): hostname of the server
        post (int): port of the server

    `Here's a sample <https://github.com/labmlai/labml/blob/master/helpers/labml_helpers/datasets/remote/test/mnist_train.py>`_
    """

    def __init__(self, name: str, host: str = "0.0.0.0", port: int = 8000):
        self.name = name
        self.port = port
        self.host = host
        self.http = urllib3.PoolManager()
        self._len = None

    def __getitem__(self, item):
        r = self.http.request('GET', f'http://{self.host}:{self.port}/{self.name}/item/{item}')
        return pickle.loads(r.data)

    def __len__(self):
        if self._len is None:
            r = self.http.request('GET', f'http://{self.host}:{self.port}/{self.name}/len')
            self._len = pickle.loads(r.data)

        return self._len


def _test():
    dataset = RemoteDataset('mnist_train')
    print(len(dataset))
    img = dataset[0]

    plt.imshow(img[0][0], cmap='gray')
    plt.show()


if __name__ == '__main__':
    _test()
